/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_CONV_H_
#define AIDGE_CORE_OPERATOR_CONV_H_

#include <array>
#include <cmath>
#include <cstddef>
#include <numeric>
#include <vector>

#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/Operator.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
enum class ConvAttr { StrideDims, DilationDims, InChannels, OutChannels, KernelDims };

template <DimIdx_t DIM>
class Conv_Op : public Operator,
                public Registrable<Conv_Op<DIM>, std::string, std::unique_ptr<OperatorImpl>(const Conv_Op<DIM> &)>,
                public StaticAttributes<ConvAttr, std::array<DimSize_t, DIM>, std::array<DimSize_t, DIM>, DimSize_t,
                                       DimSize_t, std::array<DimSize_t, DIM>> {
private:
    // FIXME: change accessibility
    std::array<std::shared_ptr<Tensor>, 3> mInputs = {std::make_shared<Tensor>(), std::make_shared<Tensor>(),
                                                      std::make_shared<Tensor>()};
    const std::shared_ptr<Tensor> mOutput = std::make_shared<Tensor>();

public:
    static constexpr const char *Type = "Conv";

    Conv_Op() = delete;

    using Attributes_ = StaticAttributes<ConvAttr, std::array<DimSize_t, DIM>, std::array<DimSize_t, DIM>,
                                             DimSize_t, DimSize_t, std::array<DimSize_t, DIM>>;
    template <ConvAttr e>
    using attr = typename Attributes_::template attr<e>;

    constexpr Conv_Op(DimSize_t in_channels,
                      DimSize_t out_channels,
                      const std::array<DimSize_t, DIM> &kernel_dims,
                      const std::array<DimSize_t, DIM> &stride_dims = create_array<DimSize_t,DIM>(1),
                      const std::array<DimSize_t, DIM> &dilation_dims = create_array<DimSize_t,DIM>(1))
        : Operator(Type),
          Attributes_(attr<ConvAttr::StrideDims>(stride_dims),
                      attr<ConvAttr::DilationDims>(dilation_dims),
                      attr<ConvAttr::InChannels>(in_channels),
                      attr<ConvAttr::OutChannels>(out_channels),
                      attr<ConvAttr::KernelDims>(kernel_dims)) {
        setDatatype(DataType::Float32);
    }

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    Conv_Op(const Conv_Op<DIM>& op)
        : Operator(Type),
          Attributes_(op),
          mOutput(std::make_shared<Tensor>(*op.mOutput))
    {
        // cpy-ctor
        setDatatype(op.mOutput->dataType());
        mImpl = op.mImpl ? Registrar<Conv_Op<DIM>>::create(mOutput->getImpl()->backend())(*this) : nullptr;
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::Conv_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<Conv_Op<DIM>>(*this);
    }

    // Data operator[](const char* inputName) override final {
    //     std::shared_ptr<Tensor> in = (strcmp(inputName, "data")) ? mInputs[0] :
    //         (strcmp(inputName, "weight") ? mInputs[1] :
    //         (strcmp(inputName, "bias") ? mInputs[2] :
    //         nullptr));
    //     assert((in!=nullptr) && "No such parameter");
    //     return *in;
    // }

    // std::shared_ptr<Conv_Op> clone() const override final {

    // }

    void associateInput(const IOIndex_t inputIdx, std::shared_ptr<Data> data) override final {
        assert(inputIdx < 3 && "operators supports only 3 inputs");
        assert(strcmp(data->type(), Tensor::Type) == 0 && "input data must be of Tensor type");

        mInputs[inputIdx] = std::dynamic_pointer_cast<Tensor>(data);
    }

    void computeOutputDims() override final {
        if (!mInputs[0]->empty()) {
            std::array<DimSize_t, DIM + 2> outputDims = {};

            for (std::size_t dim = 0; dim < this->template getAttr<ConvAttr::KernelDims>().size() ; ++dim) {
                const DimSize_t kernelExtent = this->template getAttr<ConvAttr::DilationDims>()[dim] *
                                                       (this->template getAttr<ConvAttr::KernelDims>()[dim] - 1) +
                                               1;

                outputDims[dim+2] = 1 + static_cast<DimSize_t>(
                        floor(static_cast<float>(mInputs[0]->dims()[dim+2] - kernelExtent) /
                              static_cast<float>(this->template getAttr<ConvAttr::StrideDims>()[dim])));
            }

            outputDims[1] = this->template getAttr<ConvAttr::OutChannels>();
            outputDims[0] = mInputs[0]->dims()[0];
            mOutput->resize(outputDims);
        }
    }

    bool outputDimsForwarded() const override final { return !(mOutput->empty()); }

    std::vector<std::pair<std::size_t, std::vector<DimSize_t>>> computeReceptiveField(const std::size_t firstIdx, const std::vector<DimSize_t>& outputDims, const IOIndex_t outputIdx = 0) const override {
        if (outputIdx != 0) {
            AIDGE_THROW_OR_ABORT(std::runtime_error, "Conv_Op Operator has got only one output Tensor.");
        }
        if ((outputDims.size() == (DIM+2)) && outputDimsForwarded()) {
            // Offset
            const auto outputIdxDims = mOutput->getCoord(firstIdx);
            auto inputIdxDims = outputIdxDims; // batch idx is the same
            inputIdxDims[1] = 0; // each channel is used so start with the first one

            for (DimIdx_t i = 0; i < (DIM+2); ++i) {
                if (((outputDims[i] + outputIdxDims[i]) > mOutput->dims<DIM+2>()[i]) || (outputDims[i] == 0)) {
                    AIDGE_THROW_OR_ABORT(std::runtime_error, "Given outputDim out of range for dimension %lu (%lu + %lu)", static_cast<std::size_t>(i), outputIdxDims[i], outputDims[i]);
                }
            }

            // padding is not a parameter of Conv_Op. It is handled in Pad_Op Operator
            // Width
            std::vector<DimSize_t> inputDims;
            inputDims.push_back(outputDims[0]); // same batch value
            inputDims.push_back(mInputs[0]->dims()[1]); // every input channel is used

            for (DimIdx_t i = 0; i < DIM; ++i) {
                inputDims.push_back((outputDims[2+static_cast<std::size_t>(i)] - 1)
                            * this->template getAttr<ConvAttr::StrideDims>()[static_cast<std::size_t>(i)]
                            + 1
                            + (this->template getAttr<ConvAttr::KernelDims>()[static_cast<std::size_t>(i)] - 1)
                            * this->template getAttr<ConvAttr::DilationDims>()[static_cast<std::size_t>(i)]);
                inputIdxDims[2+i] *= this->template getAttr<ConvAttr::StrideDims>()[static_cast<std::size_t>(i)];
            }
            std::vector<std::pair<std::size_t, std::vector<DimSize_t>>> res = std::vector<std::pair<std::size_t, std::vector<DimSize_t>>>();
            res.push_back(std::pair<std::size_t, std::vector<DimSize_t>>(mInputs[0]->getIdx(inputIdxDims), inputDims));
            return res;
        }
        AIDGE_THROW_OR_ABORT(std::runtime_error, "Given outputDim out of range or output dim not forwarded yet.");
    }


    inline Tensor& input(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < 3 && "operators supports only 3 inputs");
        return *(mInputs[inputIdx].get()); }
    inline Tensor& output(const IOIndex_t /*outputIdx*/) const override final { return *(mOutput.get()); }


    inline std::shared_ptr<Tensor> getInput(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < 3 && "Conv Operators supports only 3 inputs");
        return mInputs[inputIdx];
    }
    inline std::shared_ptr<Tensor> getOutput(const IOIndex_t outputIdx) const override final {
        assert((outputIdx == 0) && "Conv Operator has only 1 output");
        (void) outputIdx; // avoid unused warning
        return mOutput;
    }


    std::shared_ptr<Data> getRawInput(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < 3 && "operators supports only 3 inputs");
        return std::static_pointer_cast<Data>(mInputs[inputIdx]);
    }
    std::shared_ptr<Data> getRawOutput(const IOIndex_t outputIdx) const override final {
        assert(outputIdx == 0 && "operator supports only 1 output");
        (void) outputIdx; // avoid unused warning
        return std::static_pointer_cast<Data>(mOutput);
    }


    void setBackend(const std::string &name) override {
        mImpl = Registrar<Conv_Op<DIM>>::create(name)(*this);
        mOutput->setBackend(name);

        // FIXME: temporary workaround
        mInputs[1]->setBackend(name);
        mInputs[2]->setBackend(name);
    }

    void setDatatype(const DataType &datatype) override {
        mOutput->setDatatype(datatype);

        // FIXME: temporary workaround
        mInputs[0]->setDatatype(datatype);
        mInputs[1]->setDatatype(datatype);
        mInputs[2]->setDatatype(datatype);
    }

    inline IOIndex_t nbInputs() const noexcept override final { return 3; }
    inline IOIndex_t nbDataInputs() const noexcept override final { return 1; }
    inline IOIndex_t nbOutputs() const noexcept override final { return 1; }
    static const std::vector<std::string> getInputsName(){
        return {"data_input", "weight", "bias"};
    }
    static const std::vector<std::string> getOutputsName(){
        return {"data_output"};
    }
};

template <std::array<DimSize_t, 1>::size_type DIM>
inline std::shared_ptr<Node> Conv(DimSize_t in_channels,
                                  DimSize_t out_channels,
                                  const std::array<DimSize_t, DIM> &kernel_dims,
                                  const std::string& name = "",
                                  const std::array<DimSize_t, DIM> &stride_dims = create_array<DimSize_t,DIM>(1),
                                  const std::array<DimSize_t, DIM> &dilation_dims = create_array<DimSize_t,DIM>(1)) {
    // FIXME: properly handle default w&b initialization in every cases
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Conv, not supported");
    auto conv = std::make_shared<Node>(std::make_shared<Conv_Op<static_cast<DimIdx_t>(DIM)>>(in_channels, out_channels, kernel_dims, stride_dims, dilation_dims), name);
    // addProducer(conv, 1, append(append(kernel_dims, in_channels), out_channels), "w");
    addProducer(conv, 1, append(out_channels, append(in_channels, kernel_dims)), "w");
    addProducer(conv, 2, std::array<DimSize_t, 1>({out_channels}), "b");
    return conv;
}

// helper with C-style array instead of std::array for kernel_dims to allow automatic template DIM deduction
template <DimSize_t DIM>
inline std::shared_ptr<Node> Conv(
    DimSize_t in_channels,
    DimSize_t out_channels,
    DimSize_t const (&kernel_dims)[DIM],
    const std::string& name = "",
    const std::array<DimSize_t, DIM> &stride_dims = create_array<DimSize_t,DIM>(1),
    const std::array<DimSize_t, DIM> &dilation_dims = create_array<DimSize_t,DIM>(1)) {
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Conv, not supported");
    return Conv(in_channels, out_channels, to_array(kernel_dims), name, stride_dims, dilation_dims);
}
}  // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::ConvAttr>::data[] = {
    "StrideDims",
    "DilationDims",
    "InChannels",
    "OutChannels",
    "KernelDims"
};
}

#endif /* AIDGE_CORE_OPERATOR_CONV_H_ */
