/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_SCHEDULER_H_
#define AIDGE_SCHEDULER_H_

#include <chrono>
#include <memory>
#include <set>
#include <string>
#include <vector>

namespace Aidge {
class Node;
class GraphView;

class SequentialScheduler {
public:
    struct SchedulingElement {
        SchedulingElement(
            std::shared_ptr<Node> node_,
            std::chrono::time_point<std::chrono::high_resolution_clock> start_,
            std::chrono::time_point<std::chrono::high_resolution_clock> end_)
            : node(node_), start(start_), end(end_) {}

        std::shared_ptr<Node> node;
        std::chrono::time_point<std::chrono::high_resolution_clock> start;
        std::chrono::time_point<std::chrono::high_resolution_clock> end;
    };

    SequentialScheduler(std::shared_ptr<GraphView> graphView)
        : mGraphView(graphView)
    {
        // ctor
    };
    ~SequentialScheduler() = default;

    void generateScheduling(bool verbose = false);

    /**
     * @brief Run the provided Computational Graph with a batch of data
     */
    void forward(bool forwardDims = true, bool verbose = false);

    /**
     * @brief Save in a Markdown file the order of layers execution.
     * @param fileName Name of the generated file.
     */
    void saveSchedulingDiagram(const std::string& fileName) const;

    /**
     * @brief Return a vector of Node ordered by the order they are called by the scheduler
     *
     * @return std::vector<std::shared_ptr<Node>>
     */
    std::vector<std::shared_ptr<Node>> getStaticScheduling(){
        return mStaticSchedule;
    }

private:
    /**
     * @brief Set of layers receiving an input from currently processing layers
     *
     * @param producers Set of layers ready to run.
     * @return std::set<std::shared_ptr<Node>>
     */
    std::set<std::shared_ptr<Node>> getConsumers(const std::set<std::shared_ptr<Node>>& producers) const;

    /**
     * @brief Shared ptr to the scheduled graph view
     *
     */
    std::shared_ptr<GraphView> mGraphView;
    /**
     * @brief List of SchedulingElement (i.e: Nodes with their computation time)
     *
     */
    std::vector<SchedulingElement> mScheduling;
    /**
     * @brief List of nodes ordered by their
     *
     */
    std::vector<std::shared_ptr<Node>> mStaticSchedule;
};
} // namespace Aidge

#endif /* AIDGE_SCHEDULER_H_ */
